module Brainfuck where

import Data.Char (ord)
import Data.Maybe (mapMaybe)
import Data.Tuple (swap)

import SourceMarker
import Util (fixpoint, indent, unwordsBreaking)


data BrainfuckInstr = INC | DEC
                    | RGT | LFT
                    | LOP | POL
                    | PUT | GET
                    | BFASS SourceMarker Int String
                    | BFINFO Integer
                    deriving (Eq, Show)

assocSrcBf :: [(Char,BrainfuckInstr)]
assocSrcBf = [('+',INC), ('-',DEC)
             ,('>',RGT), ('<',LFT)
             ,('[',LOP), (']',POL)
             ,('.',PUT), (',',GET)]

assocBfSrc :: [(BrainfuckInstr,Char)]
assocBfSrc = swap <$> assocSrcBf

brainfuckSource :: [BrainfuckInstr] -> String
brainfuckSource = mapMaybe $ flip lookup assocBfSrc

-- a naive optimization
optimizeBrainfuck :: [BrainfuckInstr] -> [BrainfuckInstr]
optimizeBrainfuck = optimizeFinal . fixpoint optimize . optimizeInitial
    where optimize []           = []
          optimize (INC:DEC:is) = optimize is
          optimize (DEC:INC:is) = optimize is
          optimize (RGT:LFT:is) = optimize is
          optimize (LFT:RGT:is) = optimize is
          optimize (i:is)       = i : optimize is

          optimizeInitial (RGT:is) = optimizeInitial is
          optimizeInitial (LFT:is) = optimizeInitial is
          optimizeInitial is       = is

          optimizeFinal = reverse . o . reverse
            where o (POL:DEC:LOP:is) = o is
                  o is               = is


transpile :: (String,FilePath) -> [BrainfuckInstr] -> String
transpile (date,fp) = (header ++) . (++ footer)
                    . indent 4 . unwordsBreaking 76 . (tp <$>)

    where header = unlines [ "/* Transpiled brainfuck, generated " ++ date
                             ++ ". */"
                           , ""
                           , "#include <stdio.h>"
                           , "#include <stdlib.h>"
                           , "#include <stdint.h>"
                           , "#include <time.h>"
                           , ""
                           , "#define TAPE_BITS 16"
                           , "#define TAPE_SIZE (1 << TAPE_BITS)"
                           , "#define TAPE_MASK (TAPE_SIZE - 1)"
                           , ""
                           , "static uint8_t tape[TAPE_SIZE] = {0};"
                           , "int64_t pointer = 0;"
                           , "int64_t pointerForbidden; /*TODO*/"
                           , ""
                           , "int chr, lastPut = '\\0';"
                           , "long assertions = 0, succeeded = 0;"
                           , ""
                           , "#define ANSI_BLDINV \"\\33[1m\\33[7m\""
                           , "#define ANSI_BLD \"\\33[1m\""
                           , "#define ANSI_INV \"\\33[7m\""
                           , "#define ANSI_CLR \"\\33[0m\""
                           , "#define ANSI_TAPERUPTURE ANSI_BLD \"\\33[5m\""
                           , "#define ANSI_NONEWLINE ANSI_BLD ANSI_INV"
                           , "#define ANSI_BFDELIMITER ANSI_BLD"
                           , ""
                           , "#define WRAP(PNTR) ((PNTR) & (TAPE_MASK))"
                           , "#define CHECK_FORBIDDEN(PNTR) if ((PNTR) == pointerForbidden) { fprintf(stderr, ANSI_TAPERUPTURE \"TAPE RUPTURE\" ANSI_CLR); goto EXIT; }"
                           , "#define I ++tape[WRAP(pointer)];"
                           , "#define D --tape[WRAP(pointer)];"
                           , "#define R ++pointer; CHECK_FORBIDDEN(pointer);"
                           , "#define L --pointer; CHECK_FORBIDDEN(pointer);"
                           , "#define l while (tape[WRAP(pointer)]) {"
                           , "#define r }"
                           , "#define P putchar(lastPut = tape[WRAP(pointer)]); fflush(stdout);"
                           , "#define G chr = getchar(); tape[WRAP(pointer)] = (chr > 1) * chr;"
                           , "#define A(SM, ASS, MSG) ++assertions; if (tape[WRAP(pointer)] != (ASS)) { fprintf(stderr, \"\\nASSERTION FAILED: \\\"%s\\\"\\n\", (MSG)); fflush(stderr); } else ++succeeded;"
                           , ""
                           , ""
                           , "int main() {"
                           , "    srand(time(NULL));"
                           , "    for (size_t j = 0; j < TAPE_SIZE; j++)"
                           , "        tape[j] = rand() & 0xFF;"
                           , ""
                           , "    printf(ANSI_BFDELIMITER \"* " ++ concat [if ('a' <= c && c <= 'z') || ('A' <= c && c <= 'Z') || c `elem` ".-_/" then [c] else "\\\\" ++ show (ord c) | c <- fp] ++ " * \\n\");"
                           , "    printf(ANSI_BFDELIMITER \"{{{\\n\" ANSI_CLR);"
                           , ""
                           , "    /* What follows is transpiled brainfuck. It "
                             ++ "also contains assertions. */"
                           ]

          footer = unlines [ ""
                           , "EXIT:"
                           , "    if (lastPut != '\\n')"
                           , "        printf(ANSI_NONEWLINE \"%%\\n\" ANSI_CLR);"
                           , "    printf(ANSI_BFDELIMITER \"}}}\\n\" ANSI_CLR);"
                           , ""
                           , "    if (assertions == succeeded)"
                           , "        if (assertions == 1)"
                           , "            fprintf(stderr, \"\\33[32mThe one "
                             ++ "assertion was successful.\""
                           , "                            \"\\33[0m\\n\");"
                           , "        else"
                           , "            fprintf(stderr, \"\\33[32mAll %ld "
                             ++ "assertions were successful.\""
                           , "                            \"\\33[0m\\n\", succe"
                             ++ "eded);"
                           , "    else"
                           , "        fprintf(stderr, \"\\33[31mOnly %ld out"
                             ++ " of %ld assertion%s were \""
                           , "                        \"successful (% 6.2f%%)."
                             ++ "\\33[0m\\n\", succeeded, assertions,"
                           , "                        assertions == 1 ? \"\" : "
                             ++ "\"s\", assertions == 0 ? 100 : 100."
                           , "                        * succeeded / assertions)"
                             ++ ";"
                           , "}" ]

          tp (BFASS sm ass msg) = "A(" ++ "TODO-SM" ++ ", " ++ show ass ++ ", "
                                ++ show msg ++ ")"
          tp (BFINFO mx) = "pointerForbidden = " ++ show mx ++ " + 1;"
          tp instr = case lookup instr decoder of
              Just srcC -> srcC
              Nothing   -> ""
              where decoder = [(INC,"I"),(DEC,"D"),(RGT,"R"),(LFT,"L"),
                               (LOP,"l"),(POL,"r"),(PUT,"P"),(GET,"G")]
