{-# LANGUAGE NamedFieldPuns, MultiWayIf, LambdaCase #-}

module MacroAssembler where

import Control.Monad (zipWithM)
import Control.Monad.Trans.State.Lazy
import Data.Char (ord)
import Data.Either (isRight)
import System.Directory (makeAbsolute)

import Error
import SourceMarker
import TapeAbstraction
import Util (safeReadFile, safeTail, mapFst)


{--- Tokens -------------------------------------------------------------------}

data Token = LPAREN SourceMarker
           | RPAREN SourceMarker
           | IDENT SourceMarker String
           deriving Show


tokenize :: FilePath -> String -> Either Error [Token]
tokenize filepath source = tokenize' (SourceMarker { filepath = filepath
                                                   , character = 0
                                                   , line = 0
                                                   , column = 0
                                                   , source = source }) source

tokenize' _ "" = return []
tokenize' sm (c:cs)
        | c == '('              = (LPAREN sm :) <$> tokenize' sm' cs
        | c == ')'              = (RPAREN sm :) <$> tokenize' sm' cs
        | c == ' ' || c == '\n' = tokenize' sm' cs
        | c == '\''             = case cs of
            ""          -> err
            ('\'':_) -> err
            ('\\': '\\' :'\'':cs') -> (IDENT sm "0x5c" :) <$> tokenize' (advanceMarker' "\'\\\\\'" sm) cs'
            ('\\': '\'' :'\'':cs') -> (IDENT sm "0x27" :) <$> tokenize' (advanceMarker' "\'\\\'\'" sm) cs'
            ('\\': 'n' :'\'':cs')  -> (IDENT sm "0x0a" :) <$> tokenize' (advanceMarker' "\'\\n\'" sm) cs'
            ('\\': 'x':_)          -> Left $ Error sm "Further escape sequences are necessary."
            ('\\':_)               -> err
            (d:'\'':cs') -> (IDENT sm (show $ ord d) :) <$> tokenize' (advanceMarker' ('\'':d:"\'") sm) cs'
            _ -> err
        | c == '"'              = Left $ Error sm "Strings not yet supported."
        | c == ';'              = tokenize' (advanceMarker' comment sm) cRest
        | otherwise             = (IDENT sm ident :) <$> tokenize' (advanceMarker' ident sm) iRest
    where sm' = advanceMarker c sm

          isNotSpecial c = not $ c `elem` "() \n\""
          (ident,iRest) = span isNotSpecial (c:cs)

          (comment,cRest) = span (/= '\n') (c:cs)

          err = Left $ Error sm "Invalid character literal."


{--- Expressions --------------------------------------------------------------}

data Expr = Expr SourceMarker [Expr]
          | Ident SourceMarker String

identFromExpr :: Expr -> Maybe String
identFromExpr (Ident _ ident) = Just ident
identFromExpr _               = Nothing

instance Show Expr where
    show (Expr _ exprs) = "Expr " ++ show exprs
    show (Ident _ ident) = "Ident " ++ show ident

pretty :: Expr -> String
pretty (Expr _ exprs) = "(" ++ unwords (pretty <$> exprs) ++ ")"
pretty (Ident _ ident) = ident


parse' :: [Token] -> Either Error (Expr, [Token])
parse' [] = Left $ Error (Marker "MacroAssembler.parse'") "No expression encountered."
parse' (RPAREN sm : _) = Left $ Error sm "Expression not initialized; missing '('."
parse' (LPAREN sm : ts) = do (exprs, ts') <- prs ts
                             case ts' of
                                (RPAREN _:ts'') -> return (Expr sm exprs, ts'')
                                (t:_) -> error $ "The unthinkable happened!"
                                _ -> Left . Error sm $ "Expression not finalized; missing ')'."
    where prs :: [Token] -> Either Error ([Expr], [Token])
          prs [] = return ([], [])
          prs (RPAREN sm:ts) = return ([], RPAREN sm:ts)
          prs ts = do (expr, ts') <- parse' ts
                      mapFst (expr :) <$> prs ts'
parse' (IDENT sm ident : ts) = return (Ident sm ident, ts)

parse :: [Token] -> Either Error [Expr]
parse [] = return []
parse ts = do (expr, ts') <- parse' ts
              (expr :) <$> parse ts'


{--- Macro Assembly -----------------------------------------------------------}

data Macro = Macro { name :: String
                   , params :: Expr
                   , body :: Expr }
             deriving Show

data MacroEvaluationState = MacroEvaluationState { mesMacros :: [Macro]
                                                 , mesLocal :: Integer
                                                 , mesErrors :: [Error] }
                            deriving Show

initialMacroEvaluationState = MacroEvaluationState { mesMacros = []
                                                   , mesLocal = 0
                                                   , mesErrors = [] }

evaluationError :: Error -> State MacroEvaluationState [TapeInstr]
evaluationError err = do MacroEvaluationState { mesErrors = errs } <- get
                         modify $ \s -> s { mesErrors = errs ++ [err] }
                         return []

lookupMacro :: String -> State MacroEvaluationState (Maybe Macro)
lookupMacro macroName
    = do MacroEvaluationState { mesMacros = macros } <- get
         case filter ((macroName ==) . name) macros of
             [macro] -> return $ Just macro
             []      -> return Nothing
             _       -> do evaluationError . Error (Marker "MacroAssembler.lookupMacro") $ "Multiple macros of the same name exist."
                           return Nothing

addMacro :: Macro -> State MacroEvaluationState ()
addMacro macro = do MacroEvaluationState { mesMacros = macros } <- get
                    let macroName = name macro
                    mmacro <- lookupMacro macroName
                    case mmacro of
                        Nothing -> modify $ \s -> s { mesMacros = macro : macros }
                        _       -> do evaluationError . Error (Marker "MacroAssembler.addMacro") $ "Duplicate macro " ++ show macroName ++ "."
                                      return ()

getLocal :: State MacroEvaluationState Integer
getLocal = do MacroEvaluationState { mesLocal = local } <- get
              modify $ \s -> s { mesLocal = local+1 }
              return local

macroParamAssoc :: Macro -> Expr -> Either Error [(String, Expr)]
macroParamAssoc m@(Macro { params }) = paramAssoc params
    where paramAssoc :: Expr -> Expr -> Either Error [(String, Expr)]
          paramAssoc (Ident _ pident) expr = return [(pident, expr)]
          paramAssoc (Expr _ pexprs) (Ident sm _) = Left $ Error sm "Parameter needs to be a compound expression, not an identifier."
          paramAssoc (Expr _ pexprs) (Expr sm exprs)
              | length pexprs /= length exprs = Left . Error sm $ "Parameter count mismatch in macro " ++ show m ++ "."
              | otherwise                     = concat <$> zipWithM paramAssoc pexprs exprs

parameterMatch :: Macro -> Expr -> Bool
parameterMatch macro expr = isRight $ macroParamAssoc macro expr

applyMacro :: Integer -> Macro -> Expr -> Either Error Expr
applyMacro local macro@(Macro { body }) expr
    = do assoc <- macroParamAssoc macro expr
         return $ go assoc (localize body)
    where go :: [(String, Expr)] -> Expr -> Expr
          go assoc ident@(Ident _ idnt) = case lookup idnt assoc of
            Just expr -> expr
            Nothing   -> ident
          go assoc (Expr sm exprs) = Expr sm $ go assoc <$> exprs

          replace "" = ""
          replace ('$':cs) = show local ++ "@@@LOCAL@@@" ++ replace cs
          replace (c:cs) = c : replace cs
          localize (Ident sm ident) = Ident sm $ replace ident
          localize (Expr sm exprs) = Expr sm $ localize <$> exprs

evaluate :: Expr -> Either [Error] [TapeInstr]
evaluate expr = case runState (evaluate' expr) initialMacroEvaluationState of
    (evaluated, MacroEvaluationState { mesErrors = [] }) -> return evaluated
    (_, MacroEvaluationState { mesErrors = errs }) -> Left errs

splitNameHint :: String -> (String,String)
splitNameHint namehint = (reverse $ safeTail name', reverse hint')
    where (hint',name') = span (/= ':') $ reverse namehint


mapOnIdent :: (String -> String) -> Expr -> Expr
mapOnIdent f (Expr sm exprs) = Expr sm $ mapOnIdent f <$> exprs
mapOnIdent f (Ident sm ident) = Ident sm $ f ident


evaluate' :: Expr -> State MacroEvaluationState [TapeInstr]

evaluate' expr@(Expr sm [Ident _ "macro", Ident _ name, params, body])
    = do mmacro <- lookupMacro name
         {- TODO :: verify valid macro, i.e. params contain **different** idents -}
         case mmacro of
             Just _  -> evaluationError $ Error sm "Attempted macro re-definition."
             Nothing -> do addMacro $ Macro { name = name, params = params, body = body }
                           return []
evaluate' (Expr _ (Ident sm "macro" :_)) = evaluationError $ Error sm "Invalid macro definition."

evaluate' (Expr _ (Ident sm "comment":_)) = return []

evaluate' (Expr _ [Ident sm "derive", Ident smt template, Ident smd derivation])
    = if | templateName /= derivationName -> evaluationError . Error smd
                                           $ "Cannot derive name " ++ show derivationName ++ " from unequal name " ++ show templateName ++ "."
         | templateHint == derivationHint -> evaluationError . Error smd
                                           $ "Cannot derive equal hints."
         | otherwise -> case derive 0 templateHint derivationHint of
            Left err   -> evaluationError $ err
            Right expr -> evaluate' $ Expr loc [Ident loc "macro", Ident loc derivation, parameters, Expr loc [expr]]
    where (templateName,templateHint) = splitNameHint template
          (derivationName,derivationHint) = splitNameHint derivation
          loc = smd
          parameters = Expr loc [Ident loc $ "x" ++ show j | j <- [0..(length derivationHint)-1]]
          derive :: Integer -> String -> String -> Either Error Expr
          derive _ ""    ""    = Right $ Expr loc (Ident smt template : [Ident loc $ "$x" ++ show j | j <- [0..length derivationHint-1]])
          derive j (t:_) ""    = Left . Error smd $ "Template hints left over at index " ++ show j ++ "."
          derive j ""    (d:_) = Left . Error smd $ "Derivation hints left over at index " ++ show j ++ "."
          derive j ('z':ts) ('r':ds) = do expr <- derive (j+1) ts ds
                                          return $ Expr loc [Ident loc "let:nre", Ident loc $ "$x" ++ show j, Ident loc $ "x" ++ show j, Expr loc [expr]]
          derive j ('u':ts) ('r':ds) = derive j ('z':ts) ('r':ds)
          derive j ('r':ts) ('v':ds) = do expr <- derive (j+1) ts ds
                                          return $ Expr loc [Ident loc "let:nve", Ident loc $ "$x" ++ show j, Ident loc $ "x" ++ show j, Expr loc [expr]]
          derive j ('a':ts) ('r':'w':ds) = Left . Error smd $ "Not implemented."
          derive j (t:ts) (d:ds) | t == d    = do expr <- derive (j+1) ts ds
                                                  return $ Expr loc [Ident loc "alias", Ident loc $ "$x" ++ show j, Ident loc $ "x" ++ show j, Expr loc [expr]]
                                 | otherwise =  Left . Error smd $ "Cannot derive hints " ++ show (t:ts) ++ " to " ++ show (d:ds) ++ " at index " ++ show j ++ "."
evaluate' (Expr _ (Ident sm "derive" : _)) = evaluationError $ Error sm "Invalid derivation."

evaluate' (Expr _ [Ident sm "alias", Ident _ alias, meaning, expr]) = evaluate' $ go expr
    where go :: Expr -> Expr
          go (Expr sm exprs)  = Expr sm $ go <$> exprs
          go (Ident sm ident) | alias == ident = meaning
                              | otherwise      = Ident sm ident
evaluate' (Expr _ (Ident sm "alias" : _)) = evaluationError $ Error sm "Invalid aliasing."

evaluate' (Expr sm (Ident _ "new" : params)) = case params of
    [Ident _ ident] -> return [NEW sm ident]
    _               -> evaluationError $ Error sm "Invalid use of 'new'."

evaluate' (Expr sm (Ident _ "pnt" : params)) = case params of
    [Ident _ ident] -> return [PNT sm ident]
    _               -> evaluationError $ Error sm ("Invalid use of 'pnt', seeing " ++ show (params))

evaluate' (Expr sm (Ident sm' "num" : params)) = case params of
    [Ident sm'' ident] -> case reads ident :: [(Int,String)] of
                            [(num,[])] -> if 0 <= num && num < 256 then
                                            return [NUM sm num]
                                          else evaluationError $ Error sm'' "Exceeding number literal bounds."
                            _          -> do evaluationError . Error sm' $ "Invalid number literal " ++ show ident ++ "."
                                             evaluationError . Error sm'' $ "[[See above]] Invalid number literal " ++ show ident ++ "."
    _                  -> evaluationError $ Error sm "Invalid use of 'num'."

evaluate' (Expr sm (Ident _ "fgt" : params)) = case params of
    [Ident _ ident] -> return [FGT sm ident]
    _               -> evaluationError $ Error sm "Invalid use of 'fgt'."

evaluate' (Expr sm (Ident _ "bf" : params)) = case params of
    [Ident _ ident] -> return [BF sm ident]
    _               -> evaluationError $ Error sm "Invalid use of 'bf'."

evaluate' (Expr sm (Ident _ "ass" : params)) = case params of
    [Ident asm ass', Ident _ msg] -> case reads ass' :: [(Int,String)] of
                                        [(ass,[])] -> if 0 <= ass && ass < 256 then
                                                          return [ASS sm ass msg]
                                                      else evaluationError . Error asm $ "Assertion value out of bounds."
                                        _          -> evaluationError . Error asm $ "Invalid assertion value."
    _                             -> evaluationError $ Error sm "Invalid use of 'ass'."

evaluate' (Expr sm exprs@(Ident sm' name : values))
    = do mmacro <- lookupMacro name
         case mmacro of
            Nothing -> evaluationError . Error sm' $ "Macro not defined."
            Just macro -> do let params = Expr sm values
                             loc <- getLocal
                             case applyMacro loc macro params of
                                Left err -> evaluationError err
                                Right expr -> evaluate' expr

evaluate' (Ident sm ident) = evaluationError . Error sm $ "Stray identifier."

evaluate' (Expr st exprs) = concat <$> mapM evaluate' exprs


{--- File Inclusion (not properly implemented) --------------------------------}

include :: FilePath -> IO (Either Error [Expr])
include fp' = do fp <- makeAbsolute fp' -- TODO :: file path should be interpreted relative to the one including it
                 msrc <- safeReadFile fp
                 case msrc of
                     Nothing -> return . Left . Error (FileMarker fp) $ "Could not read file " ++ show fp' ++ "."
                     Just src -> case tokenize fp src of
                         Left err -> return . Left $ err
                         Right toks -> case parse toks of
                             Left err -> return . Left $ err
                             Right exprs -> return . Right $ exprs -- TODO :: inclusion
