{-# LANGUAGE NamedFieldPuns #-}

module TapeAbstraction where

import Control.Monad.Trans.State.Lazy
import Data.Either (partitionEithers)
import Data.List (genericReplicate)
import Data.Maybe (mapMaybe)

import Brainfuck
import Error
import SourceMarker


data TapeInstr = NEW SourceMarker String
               | PNT SourceMarker String
               | NUM SourceMarker Int
               | FGT SourceMarker String
               | BF SourceMarker String
               | ASS SourceMarker Int String
               deriving Show


data ConcretionState = ConcretionState { pointer :: Integer
                                       , cells   :: [(String,Integer)]
                                       , bounds  :: (Integer,Integer) }
initialConcretionState = ConcretionState { pointer = 0
                                         , cells   = [("ORIGIN", 0)]
                                         , bounds  = (0, 0) }

integers = (0 :) . concat $ zipWith (\x y -> [x, -y]) [1..] [1..]

unoccupiedPointer :: State ConcretionState Integer
unoccupiedPointer = do ConcretionState { pointer, cells } <- get
                       return $ head [p | p <- integers, p > 0, unoccupied cells p]
    where unoccupied [] _ = True
          unoccupied ((_,q):cells) p = q /= p && unoccupied cells p

movePointer :: Integer -> State ConcretionState [BrainfuckInstr]
movePointer p = do ConcretionState { pointer, bounds = (minBound, maxBound) } <- get
                   modify $ \s -> s { pointer = p, bounds = (min p minBound, max p maxBound) }
                   if p >= pointer then
                      return $ genericReplicate (p - pointer) RGT
                   else
                      return $ genericReplicate (pointer - p) LFT

cellPointer :: String -> State ConcretionState (Maybe Integer)
cellPointer cell = do ConcretionState { cells } <- get
                      return $ lookup cell cells


concrete :: [TapeInstr] -> Either [Error] [BrainfuckInstr]
concrete instrs = case runState (mapM concrete' instrs) initialConcretionState of
    (errBfs, ConcretionState { bounds = (_, maxTape) }) -> case partitionEithers errBfs of
        ([], bfs) -> return $ [BFINFO maxTape] ++ concat bfs
        (errs, _)  -> Left errs


concrete' :: TapeInstr -> State ConcretionState (Either Error [BrainfuckInstr])

concrete' (NEW sm cell)
    = do mp <- cellPointer cell
         case mp of
             Just _  -> return . Left . Error sm $ "Cell " ++ show cell ++ " already in use."
             Nothing -> do p <- unoccupiedPointer
                           modify $ \s -> s { cells = (cell,p) : cells s }
                           return . Right $ []

concrete' (PNT sm cell)
    = do mp <- cellPointer cell
         case mp of
             Nothing -> return . Left . Error sm $ "Cell " ++ show cell ++ " never declared."
             Just p  -> Right <$> movePointer p

concrete' (NUM sm num)
    = return . Right $ genericReplicate num INC

concrete' (FGT sm cell)
    = do mp <- cellPointer cell
         case mp of
             Nothing -> return . Left . Error sm $ "Cell " ++ show cell ++ " never declared, cannot forget."
             Just p  -> do modify $ \s -> s { cells = filter (\(c,_) -> c /= cell) (cells s) }
                           return . Right $ []

concrete' (BF _ src) = return . Right $ mapMaybe (flip lookup assocSrcBf) src
concrete' (ASS sm ass msg)
    | 0 <= ass && ass < 256 = return . Right $ [BFASS sm ass msg]
    | otherwise = return . Left . Error sm $ "Assertion value out of bounds (" ++ show ass ++ ")."
