; Jonathan Frech, 2018 to 2021
; prelude.ma -- a prelude file to define basic abstractions on top of brainfuck
;               in the brainfuck compiler project

;;; Legend ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; r....: read
; a,u..: alter, unknown
; w,b,z: write, boolean, zero
; e,v,n: expression, value, new)

;;; Tape Abstraction Instructions ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; pnt: point at a named tape cell
; bf.: embed raw brainfuck instructions
; num: add a literal to current cell
; new: allocate a new named cell
; fgt: free a named cell)

;;; Ideas ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; * A derivation akin to `tau:ra -> tau:rw` could be added to the derivation
;   mechanism, as `a -> rw` and `a -> w` do not conflict as derivations.
; * Implement four stacks instead of only three.


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Prelude Core ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(macro inc:a (x) (
    (pnt x)
    (bf +)))

(macro dec:a (x) (
    (pnt x)
    (bf -)))

(macro while:re (flg bdy) (
    (pnt flg)
    (bf [)
    bdy
    (pnt flg)
    (bf ])))

(macro clr:z (x) (
    (while:re x (
        (dec:a x)))))

(macro add:va (v x) (
    (pnt x)
    (num v)))

(macro set:wv (x v) (
    (clr:z x)
    (add:va v x)))

(macro let:ne (x bdy) (
    (new x)
    bdy
    (fgt x)))

(macro let:nve (x v bdy) (
    (let:ne x (
        (set:wv x v)
        bdy))))

(macro add:za (x y) (
    (while:re x (
        (dec:a x)
        (inc:a y)))))

(macro add2:zaa (x y z) (
    (while:re x (
        (dec:a x)
        (inc:a y)
        (inc:a z)))))

(macro add:ra (x y) (
    (let:nve $z 0 (
        (add2:zaa x y $z)
        (add:za $z x)))))

(macro cpy:rw (x y) (
    (clr:z y)
    (add:ra x y)))

(macro let:nre (x y bdy) (
    (let:ne x (
        (cpy:rw y x)
        bdy))))

; From here on, derivations are possible.


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Control Structures ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(macro if:ze (flg bdy) (
    (while:re flg (
        (clr:z flg)
        bdy))))
(derive if:ze if:re)

(macro if:zee (flg true false) (
    (let:nve $alt 1 (
        (if:ze flg (
            true
            (clr:z $alt)))
        (if:ze $alt (
            false))))))
(derive if:zee if:ree)

(macro ifnot:ze (flg bdy) (
    (if:zee flg () bdy)))
(derive ifnot:ze ifnot:re)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Logic ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(macro not:a (x) (
    (let:nve $flg 1 (
        (while:re x (
            (dec:a x)
            (clr:z $flg)))
        (add:za $flg x)))))

(macro not:rb (x b) (
    (set:wv b 1)
    (if:re x (
        (dec:a b)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Comparisons ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(macro <=:uub (x y flg) (
    (set:wv flg 1)
    (while:re x (
        (ifnot:re y (
            (clr:z flg)))
        (dec:a x)
        (dec:a y)))))
(derive <=:uub <=:rrb)

(macro >=:rrb (x y flg) (
    (<=:rrb y x flg)))

(macro >:rrb (x y flg) (
    (<=:rrb x y flg)
    (not:a flg)))

(macro <:rrb (x y flg) (
    (>:rrb y x flg)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Arithmetic ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(macro sub:za (x y) (
    (while:re x (
        (dec:a x)
        (dec:a y)))))
(derive sub:za sub:ra)

(macro sub:ar (x y) (
    (sub:ra y x)))
(derive sub:ar sub:av)

(macro add:zw (x y) (
    (clr:z y)
    (add:za x y)))

; `x` is modified to be the modulus
; calculates `d = xm/y`
(macro divmod:ara (xm y d) (
    (let:nve $flg 1 (
        ; priming such that expensive comparison only has to be called once
        (add:ra y xm)
        (dec:a d)

        (while:re $flg (
            (sub:ra y xm)
            (inc:a d)
            (>=:rrb xm y $flg)))))))

(macro divmod:arw (xm y d) (
    (clr:z d)
    (divmod:ara xm y d)))

(macro div:rrw (x y d) (
    (let:nre $xm x (
        (divmod:arw $xm y d)))))

(macro divides?:rrb (y x flg) (
    (let:nre $x x (
        (let:ne $_ (
            (divmod:ara $x y $_)
            (not:rb $x flg)))))))
(derive divides?:rrb divides?:rvb)
(derive divides?:rrb divides?:vrb)

; Undefined behaviour on `x = 0` or `x = 1`.
(macro tau:ra (x t) (
    (let:nve $d 1 (
        (while:re $d (
            (let:ne $flg (
                (divides?:rrb $d x $flg)
                (if:ze $flg (
                    (inc:a t)))))
            (inc:a $d)))))))

; Undefined behaviour on `x = 0` or `x = 1`.
(macro tau:rw (x t) (
    (clr:z t)
    (tau:ra x t)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; I/O ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(macro putc:r (x) (
    (pnt x)
    (bf .)))
(derive putc:r putc:v)

(macro newline () (
    (putc:v '\n')))

(macro printBoolean:r (x) (
    (if:ree x (
        (putc:v 't')
        (putc:v 'r')
        (putc:v 'u')
    ) (
        (putc:v 'f')
        (putc:v 'a')
        (putc:v 'l')
        (putc:v 's')
    ))
    (putc:v 'e')))

(macro printDecimal:r (x) (
    (let:nve $zero '0' (
        (add:ra x $zero)
        (putc:r $zero)))))

(macro printf%03d:u (x) (
    (let:ne $digit (
        (let:nve $hundred 100 (
            (divmod:arw x $hundred $digit)
            (printDecimal:r $digit)
            (let:nve $ten 10 (
                (divmod:arw x $ten $digit)
                (printDecimal:r $digit)
                (printDecimal:r x)))))))))
(derive printf%03d:u printf%03d:r)

(macro getc:w (x) (
    (pnt x)
    (bf ,)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Assertion ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(macro assert:rvs (x v msg) (
    (pnt x)
    (ass v msg)))

(macro assert:rv (x v) (
    (assert:rvs x v -)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Three Stacks ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; The stack is grown to the strict left of the origin:
;     {[* * 0] [* * 0] [* * 0]} ... {[buf val flg] [buf val flg] [buf val flg]} {[* * 0] [* * 0] [* * 0]} ORIGIN ...

; three stacks, therefore three steps
(macro l () (
    (bf <<<)))
(macro r () (
    (bf >>>)))

(macro L' () (
    (l) (l)))
(macro L () (
    (L') (l)))

(macro R' () (
    (r) (r)))
(macro R () (
    (r) (R')))

(macro stackInitialize () (
    (pnt ORIGIN)
    (bf <)
    (bf [-]) (l)
    (bf [-]) (l)
    (bf [-]) (l)

    (bf [-]) (l)
    (bf [-]) (l)
    (bf [-]) (l)

    (R) (R)
    (bf >)))

; positioned at the flag
(macro .buffer<- () (
    (L)
    (bf <<[-]>>)
    (R)

    (bf <<[)
        (L)
        (bf +)
        (R)
    (bf -]>>)))


(macro .stack<- () (
    (L)))

(macro .movval () (
    (bf <[-]<[>+<-]>>)))

(macro .stack-> () (
    (R)))

(macro .push () (
    (.buffer<-)
    (.stack<-)

    (bf [)
    (.buffer<-)
    (.stack<-)
    (bf ])

    (.movval)
    (bf +)

    (L)
    (bf [-])
    (R)

    (.stack<-)
    (bf [-])
    (.stack->)

    (bf [)
    (.stack->)
    (bf ])))

(macro .pushX:eer (off< off> x) (
    (cpy:rw x ORIGIN)
    (pnt ORIGIN)
    off<
    (bf [-])
    off>
    (bf [)
        off<
        (bf +)
        off>
    (bf -])

    (off<)
    (bf >>)
    (.push)
    (bf <<)
    (off>)))

(macro pushA:r (x) (
    (.pushX:eer ((l)) ((r)) x)))
(macro pushB:r (x) (
    (.pushX:eer ((l) (l)) ((r) (r)) x)))
(macro pushC:r (x) (
    (.pushX:eer ((l) (l) (l)) ((r) (r) (r)) x)))

(macro .buffer-> () (
    (R)
    (bf <<[-]>>)
    (L)
    (bf <<[)
    (R)
    (bf +)
    (L)
    (bf -]>>)))

(macro .value>>>buffer () (
    (R)
    (bf <<[-]>>)
    (L)

    (bf <[)
    (R)
    (bf <+>)
    (L)
    (bf -]>)))

(macro .pop () (
    (.stack<-)

    (bf [)
    (.stack<-)
    (bf ])
    (.stack->)
    (bf [-]) ; Could theoretically be `(bf -)`.
    (.value>>>buffer)

    (.stack->)
    (bf [)
    (.buffer->)
    (.stack->)
    (bf ])))

(macro .popX:eew (off< off> x) (
    (clr:z ORIGIN)
    (pnt ORIGIN)
    off<
    (bf >>)
    (.pop)
    (bf <<)

    (bf [)
        off>
        (bf +)
        off<
    (bf -])
    off>
    (add:zw ORIGIN x)))

(macro popA:w (x) (
    (.popX:eew ((l)) ((r)) x)))
(macro popB:w (x) (
    (.popX:eew ((l) (l)) ((r) (r)) x)))
(macro popC:w (x) (
    (.popX:eew ((l) (l) (l)) ((r) (r) (r)) x)))

(macro nonempty?X:eew (off< off> flg) (
    (clr:z ORIGIN)
    (pnt ORIGIN)

    off<
    (bf [-]>>[<<+)
        off>
        (bf +)
        off<
    (bf >>-])
    (bf <<[>>+<<-])
    off>

    (add:zw ORIGIN flg)))

(macro nonempty?A:w (flg) (
    (nonempty?X:eew ((L) (l)) ((R) (r)) flg)))
(macro nonempty?B:w (flg) (
    (nonempty?X:eew ((L) (l) (l)) ((R) (r) (r)) flg)))
(macro nonempty?C:w (flg) (
    (nonempty?X:eew ((L) (l) (l) (l)) ((R) (r) (r) (r)) flg)))

(derive pushA:r pushA:v)
